/*
 * Licensed Materials - Property of IBM                               
 * Component IBM Rational ROSE_RT BUILD : TARGET_RTS                  
 * Copyright IBM Corporation 1999, 2008. All Rights Reserved.       
 * U.S. Government Users Restricted Rights -  Use, duplication or     
 * disclosure restricted by GSA ADP  Schedule Contract with IBM Corp.
 */

#include <RTActor.h>
#include <RTController.h>
#include <RTDebugger.h>
#include <RTDiag.h>
#include <RTMain.h>
#include <RTMessage.h>
#if USE_THREADS
#include <RTMutex.h>
#endif
#include <RTQFilter.h>
#include <RTSpecialSignals.h>

#if OBSERVABLE && ! USE_THREADS

class RTDispatch_filter : public RTQFilter
{
public:
    RTS_INLINE explicit     RTDispatch_filter( int threshold );

    virtual                 ~RTDispatch_filter( void );
    virtual int             select( RTMessage & );
    virtual void            accept( RTMessage & );
    RTS_INLINE RTMessage  * result( void );

private:
    int         _threshold;
    RTMessage * _result;

    // unavailable methods
#ifndef _lint
    RTDispatch_filter( const RTDispatch_filter & );
    RTDispatch_filter & operator=( const RTDispatch_filter & );
#endif
}; //lint !e1712

RTS_INLINE RTDispatch_filter::RTDispatch_filter( int threshold )
    : RTQFilter()
    , _threshold(      threshold )
    , _result   ( (RTMessage *)0 )
{
}

RTS_INLINE RTMessage * RTDispatch_filter::result( void )
{
    return _result;
}

RTDispatch_filter::~RTDispatch_filter( void )
{
#ifdef _lint
    _result = (RTMessage *)0;
#endif
}

int RTDispatch_filter::select( RTMessage & msg )
{
    RTActor * receiver = msg.receiver();

    return receiver != (RTActor *)0
	&& receiver->designLevel() < _threshold;
}

void RTDispatch_filter::accept( RTMessage & msg )
{
    _result = &msg;
}

#endif // OBSERVABLE && ! USE_THREADS

const unsigned char RTController::topPriority[ 1 << OTRTS_NUMPRIO ] =
{
    OTRTS_NUMPRIO,
       0, 1, 0, 2, 0, 1, 0, 3, 0, 1, 0, 2, 0, 1, 0,
    4, 0, 1, 0, 2, 0, 1, 0, 3, 0, 1, 0, 2, 0, 1, 0,
    5, 0, 1, 0, 2, 0, 1, 0, 3, 0, 1, 0, 2, 0, 1, 0,
    4, 0, 1, 0, 2, 0, 1, 0, 3, 0, 1, 0, 2, 0, 1, 0,
    6, 0, 1, 0, 2, 0, 1, 0, 3, 0, 1, 0, 2, 0, 1, 0,
    4, 0, 1, 0, 2, 0, 1, 0, 3, 0, 1, 0, 2, 0, 1, 0,
    5, 0, 1, 0, 2, 0, 1, 0, 3, 0, 1, 0, 2, 0, 1, 0,
    4, 0, 1, 0, 2, 0, 1, 0, 3, 0, 1, 0, 2, 0, 1, 0
};

// Call the actor fsm to process a message of priority priorityLimit or higher.
// Return 1 if there is such a message and the debugger gives us the go-ahead
// to deliver it, 0 otherwise.

int RTController::dispatch( int priorityLimit )
{
    RTMessage  * m;
    RTMessageQ * queue;
    short        priority;

#if OBSERVABLE && ! USE_THREADS
    priorityLimit = Background;
#endif

#if USE_THREADS
    if( incomingBits != 0 )
	acceptIncoming();
#endif

    if( (priority = topPriority[ internalBits ]) > priorityLimit )
	return NoneToDeliver;

    queue = &internalQ[ priority ];

#if OTRTSDEBUG == DEBUG_NONE

    m = queue->dqHead();

#else

    bool end_required = false;

    // we ask permission to deliver user-level messages

    if( priority == Synchronous )
	m = queue->dqHead();
    else if( debugger()->step_begin( this ) != 0 )
    {
	end_required = true;
	m = queue->dqHead();
    }
    else
    {
#if USE_THREADS || ! OBSERVABLE
	return PermissionDenied;
#else
	RTDispatch_filter filter( debugger()->controlThreshold() );
	short             bits = internalBits;

	do
	{
	    if( (priority = topPriority[ bits ]) > priorityLimit )
		return PermissionDenied;

	    bits ^= 1 << priority;
	    queue = &internalQ[ priority ];
	}
	while( queue->filterOne( filter ) == 0 );

	m = filter.result();
#endif // OBSERVABLE && ! USE_THREADS
    }

#endif // OTRTSDEBUG != DEBUG_NONE

    if( queue->isEmpty() != 0 )
	internalBits ^= (short)( 1 << priority );

    if( m->receiver() == (RTActor *)0 )
    {
	switch( m->signal )
	{
	case RTDestroySignal:
	    if( RTMain::shutdown( this ) != 0 )
		enter( RTJob::finished );
	    break;
	case RTWakeupSignal:
	    // this was just supposed to wake me up: done
	    break;
	default:
	    RTDiag::panic( "unexpected control message" );
	    //lint -unreachable
	    break;
	}
#if RTS_COUNT
	counts.messagesDelivered[ priority ] += 1U;
#endif
	freeMsg( m );
    }
    else if( m->deliver() != RTMessage::Deferred )
    {
#if RTS_COUNT
	// note that deferred messages will be counted later
	counts.messagesDelivered[ priority ] += 1U;
#endif
	freeMsg( m );
    }

#if OTRTSDEBUG != DEBUG_NONE
    if( end_required )
	debugger()->step_end( this );
#endif

    return OneDelivered;
}
